# frozen_string_literal: true

require 'forwardable'

class Query
  extend Forwardable

  ATTRIBUTE_AND_VALUE_DIVIDERS = /(?<!~)=/.freeze
  VALUE_TO_LIST_DIVIDERS = /(?<!~)&/.freeze

  RepeatedFilter = Class.new(StandardError)
  Group = Class.new

  attr_reader :argv

  def initialize(argv)
    @argv = argv
  end

  def build
    @build ||= argv.map.with_object([]) do |argument, query_options|
      query_element = query_element(argument)

      if query_options.find { |el| el[:attribute] == query_element[:attribute] }
        repeated_filter(query_element[:attribute])
      end

      query_options << query_element
    end
  end

  def_delegators :build, :empty?

  private

  def query_element(argument)
    return { attribute: argument, value: Group } unless argument.match?(ATTRIBUTE_AND_VALUE_DIVIDERS)

    attribute, value = argument.split(ATTRIBUTE_AND_VALUE_DIVIDERS)

    {
      attribute: sanitize(attribute),
      value: sanitize(parse_value(value))
    }
  end

  def parse_value(value)
    return value unless value&.match?(VALUE_TO_LIST_DIVIDERS)

    value.split(VALUE_TO_LIST_DIVIDERS)
  end

  def sanitize(query_part)
    return unless query_part

    if query_part.is_a?(String)
      query_part.gsub(/~=/, '=').gsub(/~&/, '&')
    else
      query_part.map { |part| sanitize(part) }
    end
  end

  def repeated_filter(attribute)
    message = "#{attribute}: Every filter can be passed only once"
    raise RepeatedFilter, message
  end
end
